// @todo, split class into 2 - 1 that runs in a window that has a youtube player, and another that doesnt.
// also, maybe we can only inject the iframe code on google / youtube urls / domains to help save cpu.

(async function() {

'use strict';


const parentExtensionUrlBase = document.currentScript.getAttribute('parentExtensionUrlBase');
const util = await import(parentExtensionUrlBase + './yts-util.mjs');
const {leadingAndTrailingFiringThrottle, watchEx} = util;
const {TimePositionMemory} = await import(parentExtensionUrlBase + './yts-TimePositionMemory.mjs');

function YoutubeSeek() {
    // The main dom element that houses the youtube player html.
    // It's also the domain object - the YouTube proprietary player api, with methods.
    let player;

    // true after .init() was called.
    let initialized = false;

    // Enable/disable application level logging.
    const showLog = false;
    const showWindowId = false;
    const showWindowUrl = false;

    const collectUsageStats = true;

    /** @type {number} */
    let mostRecentSeek;

    const extensionInControlNameMarker = "YoutubeSeek";

    // Note that 1100 is longer than the css animation duration of 1000ms - important so that we don't interrupt the anim.
    const seekButtonAnimationDurationMillis = 1100;

    // Reference to the window object that contains the most recently active youtube player.
    // This is where we send a video seek command to when the user uses keyboard controls.
    // This variable will NOT have a value if this code is executing in the same iframe as the youtube player.
    /** @type {Window} */
    let mostRecentlyActiveWindow;

    let youtubeApiOnStateChangeListener;

    const keyCodes = {
        leftArrow: 37,
        upArrow: 38,
        rightArrow: 39,
        downArrow: 40,
        spaceBar: 32,
    };

    let seekConfig = {
        icons: {}
    };

    let isMac = false;

    // Their api defines constants like YT.PlayerState.ENDED, but I don't know how to access them via my scope.
    // So, I define my own copy of them lol.
    const YoutubePlayerState = {
        "UNSTARTED": -1,
        "ENDED": 0,
        "PLAYING": 1,
        "PAUSED": 2,
        "BUFFERING": 3,
        "VIDEO_CUED": 5
    };

    // We communicate with other windows via window.postMessage(), and these constants describe the message type.
    const MessageType = {
        WINDOW_ACTIVITY: 1,
        PLAYER_SEEK_COMMAND: 2,
        TOGGLE_PLAY_PAUSE: 3,
        TOGGLE_MUTE: 4,
        USAGE_STATS: 5,
    };

    let playerInitialized = false;

    let throttledStartLookingForPlayerDomElement = leadingAndTrailingFiringThrottle(watchEx(nonThrottledStartLookingForPlayerDomElement), 500);
    let updatePlayer = leadingAndTrailingFiringThrottle(watchEx(nonThrottledUpdatePlayer), 100);

    function startLookingForPlayerDomElement(fromMutationObs) {
        throttledStartLookingForPlayerDomElement(fromMutationObs);
    }

    function isVisible(e) {
        return !!(e.offsetWidth || e.offsetHeight || e.getClientRects().length);
    }

    function isValidVideo(video) {
        // Sometimes there's a video element that has a blank src, so we skip those.
        // They now use multiple video elems per page, even w/ duplicate #player ids!
        // We need to find the active one. They seem to hide the inactive video elements, so
        // we use that as part of our validity check.
        return !!(video && video.src && isVisible(video));
    }

    function findPlayer() {
        const players = queryAll('.html5-video-player, #movie_player').filter(player => {
            const video = player.querySelector('video');
            return isValidVideo(video);
        });
        if (players.length > 1) {
            // Multiple players.
            // If this happens, we need to fix it so we can process them both, or discriminate which one to process.
            recordUsageStats('mp');
        }
        return players[0];
    }

    /**
     * @param {string} selector
     * @param {Node?} node
     * @return {Node[]}
     */
    function queryAll(selector, node) {
        return Array.from((node || document).querySelectorAll(selector));
    }

    function queryPlayerChild(selector) {
        if (!player) {
            return null;
        }
        return player.querySelector(selector);
    }

    let numChecks = 0;
    let checkedOnce = false;
    let mutationObserverEnabled = false;
    function nonThrottledStartLookingForPlayerDomElement(fromMutationObs) {
        log(`startLookingForPlayerDomElement fromMutationObs=${fromMutationObs} playerInitialized=${playerInitialized} document.hidden=${document.hidden}`);
        // Prevent potential double init that might happen via queued up setTimeout()s.
        if (playerInitialized) {
            return;
        }

        player = findPlayer();
        if (player) {
            log(`+pf player found. checkedOnce = ${checkedOnce}, mutationObserverEnabled = ${mutationObserverEnabled}`, player);
            // @todo remove after confident this doesnt happen
            if (numChecks > 0) {
                log(`found player on attempt numChecks = ${numChecks}, fromMutationObs = ${fromMutationObs}`);
            }
            playerInitialized = true;
            log(player);
            //attachPlayerStateListener();
            attachPlayerStateListener(player);

            // Disable position indicator for now - not ready yet.
            // attachPlayerTimePositionListener(player);

            attachWindowActivityListener();
            attachResizeListener();
            scheduleRenderControls();
            // We don't want to throttle player activity that comes from finding a new player. Ideally, we could
            // fire this like we do now, but also clear the queue of the throttled player so it doesnt fire later, confusing
            // the receiver as to the proper active window.
            sendPlayerActivity();
        }

        checkedOnce = true;
        numChecks++;
    }

    /**
     * If the webpage destroys the html element, eg via document.write("<html>new doc</html>")
     * then our mutation observer will cease to update. This can happen with SPAs.
     *
     * So, this looks at the html element and makes sure our marker property is still present - if not, it attaches a new observer.
     */
    function ensureMutationObserverStillAttached() {
        const target = document.documentElement;
        if (target.youtubeSeekMutationObserverEnabled) {
            return;
        }
        target.youtubeSeekMutationObserverEnabled = true;

        function isYoutubeVideoRelatedElement(node) {
            return node &&
                (
                    node.tagName === 'VIDEO' ||
                    node.id === 'movie_player' ||
                    (node.classList && node.classList.contains('html5-video-player'))
                );
        }

        function callback(mutations) {
            // @todo, is checking for new nodes sufficient? once they have multiple video elements in the dom,
            // they likely wont add new ones, they might just update them.
            const nodesAdded = mutations.some(mutation => mutation.addedNodes.length);
            const videoChanged = mutations.some(mutation => {
                return Array.from(mutation.addedNodes).some(isYoutubeVideoRelatedElement) || Array.from(mutation.removedNodes).some(isYoutubeVideoRelatedElement);
            });
            if (nodesAdded) {
                updatePlayer();
            }
            if (videoChanged) {
                log('videop changed via MO', mutations);
            }
            if (!ensureMutationObserverStillAttached.cnt) {
                ensureMutationObserverStillAttached.cnt = 0;
            }
            ensureMutationObserverStillAttached.cnt++;
            log('mo cnt',             ensureMutationObserverStillAttached.cnt);
        }

        const cb = leadingAndTrailingFiringThrottle(watchEx(callback), 500);
        let observer = new MutationObserver(cb);
        observer.observe(target, {subtree: true, childList: true});
    }

    /**
     * Creates html nodes from a string, like innerHTML.
     *
     * @param {string} html
     * @return {Node}
     */
    function ele(html) {
        const div = document.createElement("div");
        div.innerHTML = html;
        return div.firstChild;
    }

    /**
     * This func helps schedule calls to renderControls(), so that it doesn't interrupt an animation.
     */
    let renderControlsTimerId;
    function scheduleRenderControls() {
        const timeSinceSeek = now() - mostRecentSeek;
        const recentlySeeked = timeSinceSeek < seekButtonAnimationDurationMillis;
        //console.log("time since seek", timeSinceSeek);
        if (recentlySeeked) {
            renderControlsTimerId && clearTimeout(renderControlsTimerId);
            renderControlsTimerId = setTimeout(watchEx(renderControls), seekButtonAnimationDurationMillis);
        } else {
            renderControls();
        }
    }

    /**
     * If the webpage destroys the html element, eg via document.write("<html>new doc</html>")
     * then any listeners attached to that previous document will cease to fire. This can happen with SPAs.
     *
     * So, this looks at the element and makes sure our marker property is still present - if not, it attaches a new listener.
     *
     * But, that aside, for efficiency purposes we only want to poll while the document has the user's focus,
     * so we listen to the event and turn our polling interval on or off as needed.
     *
     */
    function ensureVisibilityChangeListenerIsActive() {
        const markerKey = 'YoutubeSeekVisibilityChangeMarker';
        if (!document[markerKey]) {
            document.addEventListener('visibilitychange', watchEx(updatePlayerUpdatingPollingTaskState));
            document[markerKey] = true;
        }
    }

    function updatePlayerUpdatingPollingTaskState() {
        if (!document.hidden) {
            updatePlayerUpdatingPollingTaskState.pollingIntervalTimerId = setInterval(watchEx(() => {
                updatePlayer();
            }), 1000);
        } else {
            if (updatePlayerUpdatingPollingTaskState.pollingIntervalTimerId) {
                clearInterval(updatePlayerUpdatingPollingTaskState.pollingIntervalTimerId);
                updatePlayerUpdatingPollingTaskState.pollingIntervalTimerId = undefined;
            }
        }
    }

    function pollForPlayerUpdates() {
        ensureVisibilityChangeListenerIsActive();
        updatePlayerUpdatingPollingTaskState();
    }

    /**
     * This function would be much easier if we could just always
     * uninstall the buttons, then reinstall them in the right config / size.
     * But, we cant because this func often gets called after the user clicks a button
     * (because the player goes into the buffering state, then back to playing state, triggering our event handler), when its spinning.
     * reinstalling it interrupts the animation and feels clunky.
     *
     * So, we try to just update only when needed. But, we need to account for width & config changes.
     */
    function renderControls() {

        if (player) {
            log(`renderControls has player`);
            markExtensionInControl();

            // Some yt vids hide the controls.
            // We'll skip installing our custom controls, but still let hotkeys work.
            if (!playerHasGuiControls()) {
                recordUsageStats('noGui', 1);
                return;
            }

            installControlsWrapperIfNotPresent();
            removeFlexStyleFromChaptersContainer();
            let useTinySize;
            if (hasRoomToFitNormalControls()) {
                useTinySize = false;
                recordUsageStats('fRoom', 1);
            } else if (hasRoomToFitTinyControls()) {
                useTinySize = true;
                recordUsageStats('tRoom', 1);
            } else {
                recordUsageStats('noRoom', 1);
                log(`cant fit controls, uninstallButtons`, getInstalledButtonInfo());
                uninstallButtons();
                // When we can't fit the buttons, it's sometimes because the width is 0 because the player isn't visible. So,
                // we check and try to find another player on the page.
                updatePlayer();
                return;
            }

            // First, check if any changes to the buttons or their size is needed (add, remove, or change size).
            // If so, we will just totally uninstall + install them to make the code easier.
            const buttonInfo = getInstalledButtonInfo();
            const needsUninstall = (seekConfig.showShortRewindButton !== buttonInfo.shortRewindButtonInstalled)
                                || (seekConfig.showLongRewindButton !== buttonInfo.longRewindButtonInstalled)
                                || (seekConfig.showShortForwardButton !== buttonInfo.shortForwardButtonInstalled)
                                || (seekConfig.showLongForwardButton !== buttonInfo.longForwardButtonInstalled)
                                || (useTinySize !== buttonInfo.tinySizeUsed);

            if (needsUninstall) {
                recordUsageStats('uninst', 1);

                log(`needsUninstall`, buttonInfo, seekConfig);

                uninstallButtons();
                // Now, install what's needed.
                if (seekConfig.showShortRewindButton) {
                    installButtonIntoPlayerControls(useTinySize, true, true);
                }
                if (seekConfig.showLongRewindButton) {
                    installButtonIntoPlayerControls(useTinySize, true, false);
                }
                if (seekConfig.showLongForwardButton) {
                    installButtonIntoPlayerControls(useTinySize, false, false);
                }
                if (seekConfig.showShortForwardButton) {
                    installButtonIntoPlayerControls(useTinySize, false, true);
                }




            } else {
                // Nothing, we're configured correctly.
            }

        }
    }

    function getButton(direction, duration, controlWrapper) {
        controlWrapper = controlWrapper || getControlWrapper();
        return controlWrapper && controlWrapper.querySelector(`.yts-replay-button-${direction}.yts-replay-button-${duration}`);
    }

    function hasButton(controlWrapper, direction, duration) {
        return !!getButton(direction, duration, controlWrapper);
    }

    // @todo Would be nice if this extracted the seek durations. right now, changing the seek duration in extensions options doesn't
    // cause a button reinstall, although it will seek to the updated value properly.
    function getInstalledButtonInfo() {
        const controlWrapper = getControlWrapper();
        return {
            shortRewindButtonInstalled: hasButton(controlWrapper, 'rewind', 'short'),
            longRewindButtonInstalled: hasButton(controlWrapper, 'rewind', 'medium'),
            shortForwardButtonInstalled: hasButton(controlWrapper, 'forward', 'short'),
            longForwardButtonInstalled: hasButton(controlWrapper, 'forward', 'medium'),
            tinySizeUsed: !!controlWrapper.querySelector('.yts-replay-button.yts-xs'),
        };
    }

    const wrappedListener = watchEx(onPlayerStateChangeListener);
    function attachPlayerStateListener(player) {
        player.addEventListener("onStateChange", wrappedListener);
    }

    function removePlayerStateListener(player) {
        player.removeEventListener("onStateChange", wrappedListener);
    }

    function attachPlayerTimePositionListener(player) {
        let memory = new TimePositionMemory(player);
        memory.init();
    }

    /**
     * The player sometimes starts in a state where you need to click it before the video will start loading/playing.
     * We can't install our seek buttons/controls until after that click because the native youtube player controls aren't
     * yet shown, and thus their dimensions are not measurable by us yet.
     *
     * So, we wait for the first time the player changes to a state that's associated with the gui controls being shown,
     * and then we render our buttons.
     */
    function onPlayerStateChangeListener(playerStateCode) {
        const statesThatShowControls = [YoutubePlayerState.PAUSED, YoutubePlayerState.PLAYING];
        // const statesThatShowControls = [YoutubePlayerState.BUFFERING, YoutubePlayerState.PAUSED, YoutubePlayerState.PLAYING];
        if (statesThatShowControls.includes(playerStateCode)) {
            scheduleRenderControls();

            // We render again in a moment because it seems youtube can change the width of their controls in this
            // time period, which causes us to run out of room, pushing the floated native youtube controls down where they cant
            // be seen or used.
            setTimeout(watchEx(scheduleRenderControls), 1000);
        }
        log(`YTe event ${playerStateCode}`, arguments, player);
    }

    function attachWindowActivityListener() {
        if (!attachWindowActivityListener.attached) {
            attachWindowActivityListener.attached = true;
            // I don't see the need to send activity too often, but we don't want to throttle for too long, or
            // we might miss legitimate back and forth window focus switching.
            log('attachWindowActivityListener');
            const throttledSendPlayerActivity = leadingAndTrailingFiringThrottle(watchEx(sendPlayerActivity), 200);
            document.addEventListener("keydown", watchEx(evt => {
                throttledSendPlayerActivity();
            }));
            document.addEventListener("mousedown", watchEx(evt => {
                throttledSendPlayerActivity();
            }));
        }
    }

    function attachWindowMessageListener() {
        window.addEventListener("message", watchEx(event => {
            // Skip messages from other peoples code, & messages from our self.
            if (!event || !event.data || !event.data.type || window === event.source || event.data.messageSrcAppName !== 'yts') {
                return;
            }

            switch (event.data.type) {
                case MessageType.WINDOW_ACTIVITY:
                    mostRecentlyActiveWindow = event.source;
                    log("got player activity", event);
                    break;
                case MessageType.PLAYER_SEEK_COMMAND:
                    log("got seek command", event);
                    dispatchSeekRelative(event.data.seekValue);
                    break;
                case MessageType.TOGGLE_PLAY_PAUSE:
                    log("got toggle play/pause command", event);
                    dispatchPlayPause();
                    break;
                case MessageType.TOGGLE_MUTE:
                    log("got toggle mute command", event);
                    dispatchMuteUnMute();
                    break;
                case MessageType.USAGE_STATS:
                    // do nothing.
                    break;
                default:
                    log("unknown event type",  event);
            }

        }), false);
    }

    // youtube is going in the SPA direction, where navigating is done by just updating the dom instead of loading a new page.
    // This totally screws this extension. For now, we will try to detect this and recover
    // by looking for the updated player.
     function nonThrottledUpdatePlayer() {
        if (playerSeemsValid(player)) {
            return;
        }

        if (player) {
            let newPlayer = findPlayer();
            if (player !== newPlayer) {
                log(`found new player`, player, newPlayer);
                recordUsageStats('upd', 1);
                sendPlayerActivity();

                // Remove the event listener from the old player.
                removePlayerStateListener(player);
                // Clear the current player so our code will look for one, and find the new one.
                player = null;
                playerInitialized = false;
                startLookingForPlayerDomElement(false);
            }
        } else {
            startLookingForPlayerDomElement(false);
        }
    }

    function playerSeemsValid(player) {
        if (!player) {
            return false;
        }

        const video = player.querySelector('video');
        if (!video || !document.body.contains(video)) {
            return false;
        }

        return isValidVideo(video);
    }

    function sendMessageToTop(obj) {
        log("sendMessageToTop", obj);
        sendPostMessage(window.top, obj);
    }

    function sendSeekToActivePlayer(seekDuration) {
        if (mostRecentlyActiveWindow) {
            // Maybe we should catch exceptions here?
            // Maybe this window obj no longer exists, eg due to iframe element removal.
            sendPostMessage(mostRecentlyActiveWindow, {type: MessageType.PLAYER_SEEK_COMMAND, seekValue: seekDuration});
        }
    }

    function sendTogglePlayPauseCommandToActivePlayer() {
        if (mostRecentlyActiveWindow) {
            // Maybe we should catch exceptions here?
            // Maybe this window obj no longer exists, eg due to iframe element removal.
            sendPostMessage(mostRecentlyActiveWindow, {type: MessageType.TOGGLE_PLAY_PAUSE});
        }
    }

    function sendToggleMuteCommandToActivePlayer() {
        if (mostRecentlyActiveWindow) {
            // Maybe we should catch exceptions here?
            // Maybe this window obj no longer exists, eg due to iframe element removal.
            sendPostMessage(mostRecentlyActiveWindow, {type: MessageType.TOGGLE_MUTE});
        }
    }

    function sendPlayerActivity() {
        sendMessageToTop({type: MessageType.WINDOW_ACTIVITY});
    }

    function sendPostMessage(window, msg) {
        const m = Object.assign({}, msg, {messageSrcAppName: 'yts'});
        window.postMessage(m, "*");
    }

    function installButtonIntoPlayerControls(useTinySize, isRewind, isShortSeek) {
        const arrowDirectionWord = isRewind ? 'LeftArrow' : 'RightArrow';
        const arrowDirectionProperty = isRewind ? 'leftArrow' : 'rightArrow';
        const seekDurationWord = isShortSeek ? 'short' : 'medium';
        let hotKeyDescription;
        if (isMac) {
            hotKeyDescription = isShortSeek ? `Ctrl + Command + ${arrowDirectionWord}` : `Ctrl + Shift + ${arrowDirectionWord}`;
        } else {
            hotKeyDescription = isShortSeek ? `Ctrl + ${arrowDirectionWord}` : `Ctrl + Shift + ${arrowDirectionWord}`;
        }
        const hotKeyProperty = `ctrl${isShortSeek ? '' : '.shift'}.${arrowDirectionProperty}`;
        installReplayButton(hotKeyProperty, seekDurationWord, hotKeyDescription, isRewind, useTinySize);
    }

    function uninstallButtons() {
        queryAll(".yts-replay-button").forEach(el => {
            el.parentNode.removeChild(el);
        });
    }

    function getControlWrapper() {
        return queryPlayerChild(".yts-control-wrapper");
    }

    function installControlsWrapperIfNotPresent() {
        if (!getControlWrapper()) {
            const timeContainer = queryPlayerChild(".ytp-time-display");
            const wrapper = ele("<div class='yts-control-wrapper'></div>");
            timeContainer.parentNode.insertBefore(wrapper, timeContainer.nextSibling);
        }
    }

    /**
     * They use flex on the container which shows chapter titles, which causes it to take up all remaining space, even though it doesn't need it all to display chapter titles.
     * So, we remove the flex style as long it we can still see the float style - in this case, the floated element will
     * just shrink to fit, which will still display the titles to the user. Once we do this, our other code will now see
     * available space to display our buttons. Also, our buttons will be inserted to the left of the chapters, which will give
     * our buttons consistent location, and will allow the chapters to squeeze a bit in some cases.
     *
     * btw - very few videos use chapters.
     */
    function removeFlexStyleFromChaptersContainer() {
        const chapterContainer = queryPlayerChild(".ytp-chapter-container");
        if (chapterContainer) {
            const style = window.getComputedStyle(chapterContainer);
            // They use display:none for videos w/o any chapters. We wont modify anything unless the vid has chapters, mostly because
            // it's dangerous to change their flex style, and I'd rather only break the few videos that have chapters, if it ever breaks.
            if (style && style.display !== 'none' && style.flex) {
                chapterContainer.style.flex = '0 1 auto';
                chapterContainer.style.webkitFlex = '0 1 auto';
            }
        }
    }

    function playerHasGuiControls() {
        return queryPlayerChild(".ytp-chrome-controls");
    }

    /**
     * Our other extension also controls / seeks video elements. It's generic, meant to work on any website, opposed to this extension,
     * which is dedicated to youtube. So, if both extensions are active, we want this extension to control youtube videos
     * because it's more specialised / better for youtube. So, we mark the dom element, and the other extension looks for this mark
     * so that it wont try to control this specific player element.
     */
    function markExtensionInControl() {
        if (player) {
            const videoElem = queryPlayerChild('video');
            if (videoElem) {
                videoElem.dataset[extensionInControlNameMarker] = "1";
            }
        }
    }

    /**
     * If we have a button installed for this direction and duration of seek command, we send a click event to the button so that it animates.
     * Otherwise, we just seek the player directly.
     *
     * @param {'rewind' | 'forward'} direction
     * @param {'short' | 'medium'} duration
     * @return {boolean}
     */
    function simulateButtonClick(direction, duration) {
        const button = getButton(direction, duration);
        if (!button) {
            return false;
        }

        const event = new MouseEvent('click', {
            view: window,
            bubbles: true,
            cancelable: true
        });

        button.dispatchEvent(event);
        return true;
    }

    function installReplayButton(seekValKey, name, hotKey, isRewind, useTinySize) {
        const seekDuration = getSeekVal(seekValKey);
        const imgUrl = seekConfig.icons.replay1;
        const direction = isRewind ? "Rewind" : "Fast Forward";
        const seekDurationDisplay = seekDuration.toFixed(1);
        const imgClass = isRewind ? "" : "flip-along-vertical-line";
        let title = `${direction} ${seekDurationDisplay} seconds. Hotkey: ${hotKey}. Change this via the Extension options page (click the icon in the top right of your browser).`;
        if (!seekConfig.showSeekButtonTooltips) {
            title = '';
        }
        const button = ele(`<button class='ytp-button yts-replay-button yts-replay-button-${name} yts-replay-button-${isRewind ? 'rewind' : 'forward'} ${useTinySize ? 'yts-xs' : ''} yts-block' aria-label='${title}' title='${title}'><span class="yts-centered yts-seek-duration ${imgClass}" style='background-image: url(${imgUrl}); margin-top: -2px;'><small>33</small></span></button>`);
        getControlWrapper().appendChild(button);
        const img = button.getElementsByTagName("span")[0];
        const seekDurationDisplayElem = button.getElementsByTagName("small")[0];
        seekDurationDisplayElem.textContent = isRewind ? "-" + seekDuration : "+" + seekDuration;
        //log(seekDurationDisplayElem);
        const clickSequence = 1;
        const imgRotationInfo = {clickSequence: clickSequence, img: img, seekDurationText: seekDurationDisplayElem, rotateDirection: isRewind ? 1 : -1, useTinySize};
        setRotationStyles(imgRotationInfo);

        button.addEventListener("click", watchEx(() => {
            seekRelative(imgRotationInfo.rotateDirection * -1 * getSeekVal(seekValKey));
            setRotationStyles(imgRotationInfo);
        }));
    }

    function setRotationStyles(imgRotationInfo) {
        imgRotationInfo.clickSequence++;
        imgRotationInfo.img.style.transform = `scale(${-imgRotationInfo.rotateDirection}, 1) rotate(${360 * imgRotationInfo.clickSequence}deg`;
        imgRotationInfo.seekDurationText.style.transform = `scale(${-imgRotationInfo.rotateDirection}, 1) rotate(${360 * imgRotationInfo.clickSequence}deg) ${imgRotationInfo.useTinySize ? ' translateY(-1px)' : ''}`;
    }

    function seekRelative(offset) {
        const elem = getVideoElementFromPlayer();
        elem && seekRelativeNew(elem, offset, elem.loop, true);
        mostRecentSeek = now();
    }

    function getVideoElementFromPlayer() {
        return queryPlayerChild('video');
    }

    function now() {
        return new Date().getTime();
    }

    /**
     *
     * @param {HTMLMediaElement} media
     * @param {number} seconds
     * @param {boolean} wrap
     * @param {boolean} fastSeek
     */
    function seekRelativeNew(media, seconds, wrap, fastSeek) {
        let newPos;
        let duration = media.duration;
        let currentTime = media.currentTime;
        const isAtEnd = cmpToPrecision(duration, currentTime, 3) === 0;
        const infinitesimal = 0.001;

        if (wrap) {
            newPos = (currentTime + duration + seconds) % duration;
        } else {
            newPos = Math.max(infinitesimal, Math.min(duration - infinitesimal, currentTime + seconds));
        }

        if (fastSeek && media.fastSeek) {
            media.fastSeek(newPos);
            log(`fs`);
        } else {
            media.currentTime = newPos;
        }

        // If the user seeks to the end, the video stops playing. Even if they then seek backwards by a bit,
        // the video stays paused. But, this sucks - if were at the end and seek backwards, I think  it should start playing again.
        if (isAtEnd) {
            media.play();
        }
    }

    /**
     * Compares 2 numbers, but only using a certain number of decimal places for the comparison.
     *
     * @param a
     * @param b
     * @param decimalPlaces
     * @example cmpPrecision(1.123, 1.124, 2) === 0 && cmpPrecision(1.123, 1.124, 3) !== 0
     * @return {number}
     */
    function cmpToPrecision(a, b, decimalPlaces) {
        const pow = Math.pow(10, decimalPlaces);
        return Math.trunc(a * pow) - Math.trunc(b * pow);
    }

    /**
     * If this code is executing in the same iframe as the youtube player, then this func will call it's api to seek.
     * Otherwise, it will send a cross window communication event to tell the code in the correct iframe to seek.
     *
     * @param offset
     */
    function dispatchSeekRelative(offset) {
        if (player) {
            seekRelative(offset);
            recordUsageStats('seek', 1);
        } else if (mostRecentlyActiveWindow) {
            sendSeekToActivePlayer(offset);
        }
    }

    function dispatchPlayPause() {
        if (player) {
            togglePlayPause();
        } else if (mostRecentlyActiveWindow) {
            sendTogglePlayPauseCommandToActivePlayer();
        }
    }

    function togglePlayPause() {
        if (!player) {
            return;
        }

        // todo, maybe also allow buffering state?
        if (player.getPlayerState() === YoutubePlayerState.PLAYING) {
            player.pauseVideo();
        } else {
            player.playVideo();
        }
    }

    function dispatchMuteUnMute() {
        if (player) {
            toggleMuteUnMute();
        } else if (mostRecentlyActiveWindow) {
            sendToggleMuteCommandToActivePlayer();
        }
    }

    function toggleMuteUnMute() {
        if (!player) {
            return;
        }

        if (player.isMuted()) {
            player.unMute();
        } else {
            player.mute();
        }
    }

    function getWidthPerButton(isTinySize) {
        return isTinySize ? 18 : 36;
    }

    function widthNeededToFitControls(isTinySize) {
        const widthPerButton = getWidthPerButton(isTinySize);
        const numButtons = [
            seekConfig.showShortRewindButton,
            seekConfig.showLongRewindButton,
            seekConfig.showShortForwardButton,
            seekConfig.showLongForwardButton
        ].filter(val => !!val).length;
        return widthPerButton * numButtons;
    }

    function sumWidthOfChildren(elem) {
        return Array.from(elem.children).reduce((accum, childElem) => {
            return accum + Math.ceil(childElem.getBoundingClientRect().width);
        }, 0);
    }

    function availableWidthAmongChildren(parentElem) {
        return parentElem.clientWidth - sumWidthOfChildren(parentElem);
    }

    function widthNeededToFitNormalControls() {
        return widthNeededToFitControls(false);
    }

    function widthNeededToFitTinyControls() {
        return widthNeededToFitControls(true);
    }

    function availableWidthInChromeControlArea() {
        // The 3 children should be .ytp-left-controls and .ytp-right-controls, (and sometimes .ytp-chapter-container )
        return availableWidthAmongChildren(queryPlayerChild(".ytp-chrome-controls")) + existingCustomControlsWidth();
    }

    function availableWidthInLeftControlArea() {
        return availableWidthAmongChildren(queryPlayerChild(".ytp-left-controls")) + existingCustomControlsWidth();
    }

    function availableWidthInControlArea() {
        // log(`availableWidthInControlArea Math.max(${availableWidthInLeftControlArea()}, right=${availableWidthInChromeControlArea()}) = ${Math.max(availableWidthInLeftControlArea(), availableWidthInChromeControlArea())}`);
        // Youtube changed their layout back and forth, making us need to consider 2 possible GUI layouts they might use.
        // In one version, the .ytp-left-controls is a float, and tends to be only 200px wide. So, there's open space to the right side of it in a sibling position.
        // The other version, .ytp-left-controls uses display:flex and so will grow wide, taking up all available space in its parent.
        // When this happens, the free space should be inside of it.
        // So, we measure free space both inside of it, and inside its parent, and use the larger one scenario.
        // But, in either case, we always inject our custom controls to be a child of .ytp-left-controls.
        return Math.max(availableWidthInLeftControlArea(), availableWidthInChromeControlArea());
    }

    // TODO check if left has any room, according to its children. this is for the flex case, because the left outside dims will be full width.
    // also check how much room .ytp-chrome-controls has.

    function existingCustomControlsWidth() {
        const wrapper = getControlWrapper();
        return wrapper ? wrapper.clientWidth : 0;
    }

    function hasRoomToFitNormalControls() {
        log(`hasRoomToFitNormalControls needed=${widthNeededToFitNormalControls()} < avail=${availableWidthInControlArea()}`);
        return widthNeededToFitNormalControls() < availableWidthInControlArea();
    }

    function hasRoomToFitTinyControls() {
        log(`hasRoomToFitTinyControls needed=${widthNeededToFitTinyControls()} < avail=${availableWidthInControlArea()}`);
        return widthNeededToFitTinyControls() < availableWidthInControlArea();
    }

    function getSeekVal(key) {
        return parseFloat(seekConfig[key] || 1);
    }

    function setupKeyListeners() {
        function recordHotkeyUsage() {
            recordUsageStats('hk', 1);
        }

        document.addEventListener("keydown", watchEx(evt => {

            if (evt.ctrlKey) {
                if (evt.shiftKey) {
                    if (evt.which === keyCodes.leftArrow) {
                        recordHotkeyUsage();
                        if (!simulateButtonClick('rewind', 'medium')) {
                            dispatchSeekRelative(-getSeekVal("ctrl.shift.leftArrow"));
                        }
                    } else if (evt.which === keyCodes.rightArrow) {
                        recordHotkeyUsage();
                        if (!simulateButtonClick('forward', 'medium')) {
                            dispatchSeekRelative(getSeekVal("ctrl.shift.rightArrow"));
                        }
                    } else if (evt.which === keyCodes.downArrow) {
                        recordHotkeyUsage();
                        // Seek to start: this will figure out the max/min position.
                        dispatchSeekRelative(-1000000);
                    } else if (evt.which === keyCodes.upArrow) {
                        recordHotkeyUsage();
                        // Seek to end: this will figure out the max/min position.
                        dispatchSeekRelative(1000000);
                    } else if (evt.which === keyCodes.spaceBar) {
                        recordHotkeyUsage();
                        dispatchPlayPause();
                    }
                } else {
                    if (evt.which === keyCodes.leftArrow) {
                        recordHotkeyUsage();
                        if (!simulateButtonClick('rewind', 'short')) {
                            dispatchSeekRelative(-getSeekVal("ctrl.leftArrow"));
                        }
                    } else if (evt.which === keyCodes.rightArrow) {
                        recordHotkeyUsage();
                        if (!simulateButtonClick('forward', 'short')) {
                            dispatchSeekRelative(getSeekVal("ctrl.rightArrow"));
                        }
                    } else if (evt.which === keyCodes.spaceBar) {
                        recordHotkeyUsage();

                        if (evt.altKey) {
                            dispatchMuteUnMute();
                        } else {
                            dispatchPlayPause();
                        }
                    }
                }
            }

        }));
    }

    /**
     * When the window is resized, the youtube player width adjusts accordingly, and we may no longer have enough room to
     * fit our controls in the youtube player, so we need to check. Otherwise, we will push
     * some of the native controls out of view.
     *
     * The resize event can happen rapidly, so we need to throttle it.
     */
    function attachResizeListener() {
        if (!attachResizeListener.attached) {
            attachResizeListener.attached = true;
            const renderFunc = leadingAndTrailingFiringThrottle(watchEx(renderControls), 200);
            window.addEventListener("resize", evt => {
                log('resize');
                renderFunc();
                // We re-render in a moment because sometimes measurements aren't correct, resulting in the wrong size buttons being displayed.
                // I think this is because a resize is a very resource intensive time, and so waiting a bit gives things time to settle down, allowing
                // the browser to get caught up on executing code or rendering / resizing etc...
                setTimeout(renderFunc, 1000);
            }, false);
        }
    }

    function isTopWindow() {
        return window === window.top;
    }

    function currentDomainIsYoutube() {
        if (currentDomainIsYoutube.cachedValue === undefined) {
            currentDomainIsYoutube.cachedValue = window.location.hostname.endsWith('youtube.com');
        }

        return currentDomainIsYoutube.cachedValue;
    }

    function log() {
        if (!showLog) {
            return;
        }

        // This id is helpful to differentiate which window a log event is coming from - there's often many different windows involved when
        // running this extension. As such, a separate copy of this code runs in each window.
        window.youtubeSeekWindowId = window.youtubeSeekWindowId || (new Date().getTime() - 1483925683061 - 265597);

        let logIdHelper = isTopWindow() ? "top" : "sub";
        if (showWindowId) {
            logIdHelper += window.youtubeSeekWindowId;
        }
        if (showWindowUrl) {
            logIdHelper += location.href;
        }

        const args = [logIdHelper].concat(Array.from(arguments));
        console.log.apply(console, args);
    }

    // let send = cusTh
    let throttledSendAccumulatedUsageStats = leadingAndTrailingFiringThrottle(watchEx(sendAccumulatedUsageStats), 2000);
    let accumulatedUsageStats = {};
    function sendAccumulatedUsageStats() {
        collectUsageStats && sendMessageToTop({type: MessageType.USAGE_STATS, usageStats: accumulatedUsageStats});

        // Now that we've sent it, clear out the aggregated values.
        Object.keys(accumulatedUsageStats).forEach(key => {
            delete accumulatedUsageStats[key];
        });
    }

    /**
     * Send the key + value to our content script in the parent window, where we can record the usage stats into chromes storage.
     *
     * For performance, we want to reduce how many window.postMessage() and resulting chrome storage calls we make.
     * So, this func will aggregate the usage stats, and send 1 big message every so often.
     *
     * @param key
     * @param incrementalValue
     */
    function recordUsageStats(key, incrementalValue) {
        accumulatedUsageStats[key] = accumulatedUsageStats[key] || 0;
        accumulatedUsageStats[key] += incrementalValue;

        collectUsageStats && throttledSendAccumulatedUsageStats();
    }

    this.init = watchEx(function() {
        // We still listen for keys and messages on non-youtube.com pages because this code needs to run in
        // parent pages that embed the youtube player via iframe. This allows us to send hotkeys to the child iframe
        // even if the iframe doesnt currently have focus.
        setupKeyListeners();
        attachWindowMessageListener();

        // We need to find a better way to detect player changes, but youtube changes their code so often
        // that it's always breaking us. So, for now, on youtube.com pages we will just poll and look for an updated
        // player every few seconds. The player seems to change when the user navigates within youtube, like going between a
        // video page and a channel page.
        if (currentDomainIsYoutube()) {
            updatePlayer();
            pollForPlayerUpdates();
            recordUsageStats('init', 1);
        }

        log(`init yts`);
    });

    this.recordError = function() {
        recordUsageStats('err', 1);
    };

    this.setConfig = watchEx(function(cfg) {
        log(cfg);
        seekConfig = cfg;
        isMac = seekConfig.os === 'mac';
        if (!initialized) {
            this.init();
            initialized = true;
        }
        // Need to update the controls so they use the newest config values.
        scheduleRenderControls();
    });

    (function construct() {
        // Expose the player as a read only attribute.
        Object.defineProperty(this, 'player', {
            get: () => player,
            enumerable: true
        });
    }).call(this);
}
YoutubeSeek.numErrors = 0;

const ys = new YoutubeSeek();
// We call init only after we receive the config via setConfig().
// ys.init();
window.YoutubeSeek = ys;

})();

