/**
 * This script injects our js and css into the webpage after it loads. It also
 * serves as a communication point between the webpage and the extension, allowing the
 * webpage to update if the user changes their settings.
 */


const showLog = false;

const MessageType = {
    WINDOW_ACTIVITY: 1,
    PLAYER_SEEK_COMMAND: 2,
    TOGGLE_PLAY_PAUSE: 3,
    TOGGLE_MUTE: 4,
    USAGE_STATS: 5,
};

(async () => {
    const exports = await import(chrome.runtime.getURL('./saved-settings-misc.mjs'));
    const getSavedOptions = exports.getSavedOptions;
    const updateSynchronizer = new ChromeStorageUpdateSynchronizer(chrome.storage.local);

    injectFile(chrome.extension.getURL('yts-controls.js'), "script").then(() => {
        setTimeout(sendUpdatedConfigToWebPage, 1000);
    });

    // injectFile(chrome.extension.getURL('yts-TimePositionMemory.mjs'), "script", {attributes: {type: "module"}}).then(() => {
    //     setTimeout(sendUpdatedConfigToWebPage, 1100);
    // });
    
    // When the settings/config changes, we inject it into the webpage so it can update without the user needing to reload the page.
    chrome.storage.onChanged.addListener(function(changes, namespace) {
        // We only need to update the page if certain config changed. We update the usageStats key very often,
        // but the page doesn't need to be updated for that, so we only update if something else changed.
        const keys = Object.keys(changes);
        const onlyUsageStatsChanged = keys.length === 1 && keys.includes('usageStats');
        if (!onlyUsageStatsChanged) {
            log('######################################## usageStats');
            sendUpdatedConfigToWebPage();
        }
    });

    window.addEventListener("message", function(event) {
        // Skip messages from other peoples code.
        if (!event || !event.data || !event.data.type || event.data.messageSrcAppName !== 'yts') {
            return;
        }

        const d = event.data;
        if (d && d.type === MessageType.USAGE_STATS && d.usageStats) {
            log('USAGE_STATS update', d);
            updateLocalStorage(d.usageStats);
        }
    }, false);

    function isTopWindow() {
        return window === window.top;
    }

    function log() {
        const args = [isTopWindow() ? "top" : "sub", location.href].concat(arguments);
        showLog && console.log.apply(console, args);
    }

    function addIconUrls(config) {
        // todo, maybe can use predefined constants to get the extension id instead of this junk
        // https://developer.chrome.com/extensions/i18n#overview-predefined @@extension_id
        config.icons = {};
        config.icons.replay1 = chrome.extension.getURL('icons/replay1.svg');
        return config;
    }

    function getIconUrls() {
        // todo, maybe can use predefined constants to get the extension id instead of this junk
        // https://developer.chrome.com/extensions/i18n#overview-predefined @@extension_id
        const config = {icons: {
            replay1: chrome.extension.getURL('icons/replay1.svg')
        }};
        return config;
    }

    function sendUpdatedConfigToWebPage() {
        Promise.all([getSavedOptions(), getSavedLocalStorage()]).then(values => {
            let syncStorage = values[0];
            let localStorage = values[1];
            let config = Object.assign({}, syncStorage, localStorage, getIconUrls());

            // We inject a script using a data url to achieve a dynamic script.
            // Need to do this junk because of chrome content script isolated environment.
            injectFile(scriptDataUrl(scriptContentCallback(config)), "script", true);
        });
    }

    function removeElement(ele) {
        if (ele && ele.parentNode) {
            ele.parentNode.removeChild(ele);
        }
    }

    function removeElementAsync(ele) {
        setTimeout(() => {
            removeElement(ele);
        }, 15);
    }

    function injectFile(url, type, options = {}) {
        return new Promise(function (resolve, reject) {
            const elem = document.createElement(type);
            switch (type) {
                case 'link':
                    elem.href = url;
                    elem.rel  = 'stylesheet';
                    elem.type = 'text/css';
                    break;
                case 'script':
                    elem.src = url;
                    break;
                default:
                    throw new Error("Unhandled switch case:" + type);
            }

            if (options.attributes) {
                Object.keys(options.attributes).forEach(key => {
                    elem.setAttribute(key, options.attributes[key]);
                });
            }

            // extension urls look like chrome-extension://fkphmjblelljkbblhlnlhnjbcofcgeka/yts-controls.mjs
            // We want to extract the part up until the path.
            const matches = url.match(/^(chrome-extension:\/\/\w{10,}\/)/i);
            if (matches && matches[1]) {
                elem.parentExtensionUrlBase = matches[1];
                elem.setAttribute('parentExtensionUrlBase', matches[1]);
            }

            elem.onload = function() {
                resolve();
                if (options.autoRemoveElement) {
                    removeElementAsync(elem);
                }
            };
            elem.onerror = () => {
                reject();
                removeElementAsync(elem);
            };

            (document.head || document.documentElement).appendChild(elem);
        });
    }

    function scriptDataUrl(scriptContents) {
        return "data:text/javascript;base64," + btoa(scriptContents);
    }

    function scriptContentCallback(configArg) {
        return `window.YoutubeSeek.setConfig(` + JSON.stringify(configArg) + ")";
    }

    function updateLocalStorage(incrementalKeyValues) {
        updateSynchronizer.updateKeys(['usageStats'], localVals => {
            localVals.usageStats = localVals.usageStats || {};

            Object.entries(incrementalKeyValues).forEach(pair => {
                let [key, incrementalValue] = pair;
                localVals.usageStats[key] = localVals.usageStats[key] || 0;
                localVals.usageStats[key] += incrementalValue;
            });

            return new Promise(resolve => {
                chrome.storage.local.set(localVals, resolve);
            });
        });
    }

    function getSavedLocalStorage() {
        return new Promise(resolve => {
            chrome.storage.local.get({os: 'n/a'}, resolve);
        });
    }

})();




